function [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,Pt,R2Z1,R2Z2] = mbcomp(Z1,Z2,Rz1,Rz2,R,a1,a2,conv);
% [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,Pt,R2Z1,R2Z2] = mbcomp(Z1,Z2,Rz1,Rz2,R,a1,a2,conv);
% 
% Twoway multiblock component model of matrices Z1 and Z2.
% The number of components can be selected different for the predictor blocks.
% After convergence all score matrices are orthogonalized.
% 
% Minimize f(Wz1,Wz2,W)= a1||Z1-Z1Wz1Pz1'||^2 + 
%                        a2||Z2-Z2Wz2Pz2'||^2 +
%                        a3||T-TWPt'||^2  
%
% Input: Z1      (I x Jz1)
%        Z2      (I x Jz2)
%        Rz1:    number of components of block Z1 (set Rz1=0 if it should not be used)
%        Rz2:    number of components of block Z2
%        R:      number of final components
%        a1:     weight of Z1-block (between 0 and 1)
%        a2:     weight of Z2-block (between 0 and 1)
%        conv:   convergence criterion (usually 1E-6)
%
% Output: Wz1:   weight vectors for Z1 (Jz1 x Rz1)
%         Wz2:   weight vectors for Z2 (Jz2 x Rz2)
%         W  :   weight vectors for Tsup (Rz1+Rz2 x R)
%         Tz1:   scores of Z1 (I x Jz1)
%         Tz2:   scores of Z2 (I x Jz2)
%         Tsup:  scores to predict Y (I x R)
%         Pz1:   loadings of Z (Rz1 x Jz1)
%         Pz2:   loadings of X (Rz2 x Jz2)
%         Pt:    loadings of T (Rz1+Rz2 x R)
%         R2Z1, R2Z2: explained amount of variation in the blocks (between 0 and 1)
%
% Uses: ssq.m, ed.m
%
% Johan Westerhuis, 25/10/99 (westerhuis@its.chem.uva.nl)
%   ==========================================================================
%   Copyright 2005 Biosystems Data Analysis Group ; Universiteit van Amsterdam
%   ==========================================================================

prntlag = 1000;

% Sizes of matrices
[I,Jz1] = size(Z1);
[I,Jz2] = size(Z2);

% Define a3
a3=1-a1-a2;

% Calculate starting values of Wz1, Wz2, and W
% First Wz1
if I>Jz1
   [P,D,Q]  = svd(Z1,0);
   Wz1  = Q(:,1:Rz1)/D(1:Rz1,1:Rz1);
else
   [P,D]    = ed(Z1*Z1');
end;
Tz1 = P(:,1:Rz1);

% Now Wz2
if I>Jz2
   [P,D,Q]  = svd(Z2,0);
   Wz2  = Q(:,1:Rz2)/D(1:Rz2,1:Rz2);
else
   [P,D]    = ed(Z2*Z2');
end;
Tz2 = P(:,1:Rz2);

T   = [Tz1 Tz2];

% And W
if I    > Rz1+Rz2
   [P,D,Q]  = svd(T,0);
   W    = Q(:,1:R)/D(1:R,1:R);
else
   [P,D]    = ed(T*T');
end;
Tsup    = P(:,1:R);

% Calculate starting values of Pz1 and Pz2
Pz1 = (pinv(Tz1)*Z1)';  % at start all T's are orthonormal
Pz2 = (pinv(Tz2)*Z2)';  
Pt    = (pinv(Tsup)*T)';

% Calculate start loss functions
lossZ1 = ssq(Z1 - Tz1*Pz1');
lossZ2 = ssq(Z2 - Tz2*Pz2');
lossT  = ssq(T  - Tsup*Pt');
f      = a1*lossZ1 + a2*lossZ2 + a3*lossT;
ff     = f;
% fprintf('function value at start is %12.8f \n',f);

if I    < Jz1,    Wz1   = pinv(Z1)*Tz1; end
if I    < Jz2,    Wz2   = pinv(Z2)*Tz2; end
if I    < Rz1+Rz2, W    = pinv(T)*Tsup; end 

% Start iterations
fold    = f+2*conv*f;
iter    = 0;

H0    = zeros(I,Rz1+Rz2);
H1    = eye(Rz1+Rz2);
while fold-f>conv*f | iter<2
   fold   = f;
   iter   = iter+1;
   
   % Update Tz1 and Wz1. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(a3) * Tz2*(H1(Rz1+1:Rz1+Rz2,:) - H2(Rz1+1:Rz1+Rz2,:));
   F    = [sqrt(a1)*Z1,  H0-H3];
   G    = [sqrt(a1)*Pz1',sqrt(a3)*(H1(1:Rz1,:) - H2(1:Rz1,:))];
   
   if I > Jz1
      Wz1   = Z1\F/G;
      Tz1   = Z1*Wz1;
   else
      Tz1   = F/G;
   end;
   
   T    = [Tz1 Tz2];
   Tsup = T*W;
   
   % Update Pz1 and Pt
   Pz1  =(pinv(Tz1)*Z1)';
   pinv_Tsup = pinv(Tsup);
   Pt   =(pinv_Tsup*T)';
   
   % Update Tz2 and Wz2. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(a3) * Tz1*(H1(1:Rz1,:) - H2(1:Rz1,:)); 
   F    = [sqrt(a2)*Z2,   H0-H3];
   G    = [sqrt(a2)*Pz2', sqrt(a3)*(H1(Rz1+1:Rz1+Rz2,:) - H2(Rz1+1:Rz1+Rz2,:))];
   
   if I  > Jz2
      Wz2   = Z2\F/G;
      Tz2   = Z2*Wz2;
   else
      Tz2   = F/G;
   end;
   
   T    = [Tz1 Tz2];
   Tsup = T*W;
   
   % Update Pz2 and Pt
   Pz2  = (pinv(Tz2)*Z2)';
   pinv_Tsup=pinv(Tsup);
   Pt   = (pinv_Tsup*T)';
   
   % Update T and W. Define auxiliary matrices F and G
   % minW ||T-TWPt||^2
   W    = pinv(Pt');
   Tsup = T*W;
   
   % Update Pt
   pinv_Tsup = pinv(Tsup);
   Pt        =  (pinv_Tsup*T)';
   
   % Calculate new f value
   lossZ1   = ssq(Z1 - Tz1*Pz1');
   lossZ2   = ssq(Z2 - Tz2*Pz2');
   lossT    = ssq(T  - Tsup*Pt');
   f        = a1*lossZ1 + a2*lossZ2 + a3*lossT;
   if ff < f, 
      error('loss function is increasing')
   end
   ff = f;
   
   if rem(iter,prntlag)==0
      fprintf('f= %12.8f after %g iters; lossZ=%10.6f lossX=%10.6f',f,iter,lossZ1,lossZ2);
   end;
end;

% Orthogonalizing Tz1 and Tz2;
[Tz1,qr1]  = qr(Tz1,0);
Wz1        = Wz1 * inv(qr1);
Pz1        = Pz1 * qr1';
W(1:Rz1,:) = qr1 * W(1:Rz1,:);

[Tz2,qr2]  = qr(Tz2,0);
Wz2      = Wz2 * inv(qr2);
Pz2        = Pz2 * qr2';
W(Rz1+1:Rz1+Rz2,:) = qr2 * W(Rz1+1:Rz1+Rz2,:);

% Orthogonalization of Tsup
[Tsup,qrsup]  = qr(Tsup,0);
W          = W * inv(qrsup);
Pt         = Pt * qrsup';

if I    < Jz1,    Wz1   = pinv(Z1)*Tz1; end
if I    < Jz2,    Wz2   = pinv(Z2)*Tz2; end
if I    < Rz1+Rz2, W    = pinv(T)*Tsup; end 

% Give percentage explained using blockscores(1) or superscores(2)
scores = 2;
switch scores
case 1
   % explained by block scores
   % disp('explained by block scores')
   R2Z1 =(1 - ssq(Z1 - Tz1*Pz1') / ssq(Z1));
   R2Z2 =(1 - ssq(Z2 - Tz2*Pz2') / ssq(Z2));
   
case 2
   % Explained by super scores
   % disp('explained by super scores')
   P1   = (pinv(Tsup)*Z1)';
   P2      = (pinv(Tsup)*Z2)';
   R2Z1 =(1 - ssq(Z1 - Tsup*P1') / ssq(Z1));
   R2Z2 =(1 - ssq(Z2 - Tsup*P2') / ssq(Z2));
end