function [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,B,C,H,Py,R2Z1,R2Z2,R2Y]=mbmw(Z1,Z2,Jz2,Kz2,Y,R1,RA,RB,RC,R,k1,k2,c,flag,conv);
% [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,B,C,H,Py,R2Z1,R2Z2,R2Y]=mbmw(Z1,Z2,Jz2,Kz2,Y,R1,RA,RB,RC,R,k1,k2,c,flag,conv);
% 
% Multiway Multiblock covariates regression. Matrix Z1 and threeway array Z2 predict Y.
% A Tucker3 structure is applied on Z2.
% R1, RA and R can be chosen differently.
% 
% Minimize f(Wz1,Wz2)  = a1||Z1-Z1Wz1Pz1'||^2 + 
%                        a2||Z2-Z2Wz2H(C'kronB')||^2 +
%                        aT||T-TWPt'||^2 +
%                        aY||Y-TWPy'||^2
%
% Input: 
%    Z1(IxJz1)         Predictor block 1
%    Z2(Ix(Jz2*Kz2))   Predictor block 2 unfolded with batch direction maintained
%    Jz2               number of variables in process variables direction
%    Kz2               number of variables in time direction
%    Y(IxJy)           Response block
%    R1                number of components for Z1
%    RA                number of components for Z2 in the batch direction
%    RB                number of components for Z2 in the process variable direction
%    RC                number of components for Z2 in the time direction
%    R                 number of final components
%    k1                ratio of a1/aY
%    k2                ratio of a2/aY
%    c                 ratio of aT/aY
%    flag              %Expl using blockscores (1) or superscores (2)
%    conv              convergence criterion (usually 1E-6)
%
% Output: 
%    Wz1(Jz1xR1)       weight vectors for Z1
%    Wz2((Jz2*Kz2)xRA) weight vectors for Z2
%    W((R1+Ra)xR       weight vectors for T)
%    Tz1(IxR)          block scores of Z1
%    Tz2(IxR)          block scores of Z2
%    Tsup(IxR)         super scores to predict Y
%    Pz1(Jz1xR1)       loadings of Z1
%    Pz2((Jz2xKz2)xRA) loadings of Z2 (Pz2=H(C'kronB'))
%    B(Jz2xRB)         loadings of Z2 in process variables direction
%    C(Kz2xRC)         loadings of Z2 in time direction
%    H(Rx(RB*RC))      core array
%    Py(JyxR)          loadings of Y
%    R2Z1              explained variation of Z1
%    R2Z2              explained variation of Z2
%    R2Y               explained variation of Y
%
% Other variables:
%    lossZ1
%    lossZ2
%    lossT
%    lossY
%    f

% Uses: ssq.m, ed.m, permnew.m
%
% Johan Westerhuis, UVA-PAC (email: westerhuis@its.chem.uva.nl)
%   ==========================================================================
%   Copyright 2005 Biosystems Data Analysis Group ; Universiteit van Amsterdam
%   ==========================================================================

% Latest version October 26, 1999

prntlag = 50;

% Sizes of matrices
[I,Jz1] = size(Z1);
[I,JK]  = size(Z2);   % JK = Jz2*Kz2
[I,Jy]  = size(Y);

% Define alpha values
aY = 1/(1+k1+k2+c); %aY + caY + k1aY + k2aY = 1
aT = c*aY;          %1  + c   + k1   + k2   = 1/aY
a1 = k1*aY;
a2 = k2*aY;

% Calculate starting values of Wz1, Wz2
% First Wz1
if I>Jz1
   [P,D,Q]   = svd(Z1,0);
   Wz1       = Q(:,1:R1)/D(1:R1,1:R1);
else
   [P,D]     = ed(Z1*Z1');
end;
Tz1      = P(:,1:R1);

% Now Wz2
if I > Jz2*Kz2
   [P,D,Q]  = svd(Z2,0);
   Wz2      = Q(:,1:RA)/D(1:RA,1:RA);
else
   [P,D]    = ed(Z2*Z2');
end;
Tz2      = P(:,1:RA);

T        = [Tz1 Tz2];

% And W
if I    > R1+RA
   [P,D,Q]   = svd(T,0);
   W         = Q(:,1:R)/D(1:R,1:R);
else
   [P,D]     = ed(T*T');
end;
Tsup       = P(:,1:R);

% Update Pz1, Pz2 and Py;
Pz1   = (pinv(Tz1)*Z1)';
Py    = (pinv(Tsup)*Y)';  
Pt    = (pinv(Tsup)*T)';

% Calculate starting values for B, C and H
% B
XX = permnew(Z2,I,Jz2,Kz2);
if Jz2 > I*Kz2
   [P,Dd,Q] = svd(XX,0);
else
   [P,Dd]   = ed(XX*XX');
end;
B = P(:,1:RB);

% C
XX = permnew(Z2,Jz2,Kz2,I);
if Kz2 > I*Jz2
   [P,Dd,Q] = svd(XX,0);
else
   [P,Dd]   = ed(XX*XX');
end;
C = P(:,1:RC);

% H
X1     = permnew(Z2,I,Jz2,Kz2);
X1     = permnew(B'*X1,RB,Kz2,I);
X1     = permnew(C'*X1,RC,I,RB);
H      = pinv(Tz2)*X1;

% Calculate Pz2
Pz2    = permnew(H,RA,RB,RC);
Pz2    = permnew(B*Pz2,Jz2,RC,RA);
Pz2    = permnew(C*Pz2,Kz2,RA,Jz2);
Pz2    = Pz2';

% Calculate start loss functions
lossZ1 = ssq(Z1 - Tz1*Pz1');
lossZ2 = ssq(Z2 - Tz2*Pz2');
lossT  = ssq(T  - Tsup*Pt');
lossY  = ssq(Y  - Tsup*Py');
f      = a1*lossZ1 + a2*lossZ2 + aT*lossT + aY*lossY;
ff     = f;

if I    < Jz1,    Wz1   = pinv(Z1)*Tz1; end
if I    < Jz2,    Wz2   = pinv(Z2)*Tz2; end
if I    < R1+RA,   W    = pinv(T)*Tsup; end 

% Start iterations
fold       = f + 2*conv*f;
iter       = 0;

H0  = zeros(I,R1+RA);
H1  = eye(R1+RA);

while fold-f>conv*f | iter<2
   fold   = f;
   iter   = iter+1;
   
   % Update Tz1 and Wz1. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(aT) * Tz2*(H1(R1+1:R1+RA,:) - H2(R1+1:R1+RA,:));
   F  = [sqrt(a1)*Z1,  H0-H3, sqrt(aY)*(Y - Tz2*W(R1+1:R1+RA,:)*Py')];
   G  = [sqrt(a1)*Pz1',sqrt(aT)*(H1(1:R1,:) - H2(1:R1,:)),  sqrt(aY)*W(1:R1,:)*Py'];
   
   % Update Tz1 and Wz1. Define auxiliary matrices F and G   
   if I > Jz1
      Wz1     = Z1\F/G;
      Tz1     = Z1*Wz1;
   else
      Tz1     = F/G;
   end;
   
   T    = [Tz1 Tz2];
   Tsup = T*W;
   
   % Update Pz1, Pt and Py
   Pz1  =(pinv(Tz1)*Z1)';
   pinv_Tsup = pinv(Tsup);
   Py   =(pinv_Tsup*Y)';
   Pt   =(pinv_Tsup*T)';
   
   % Update Tz2 and Wz2. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(aT) * Tz1*(H1(1:R1,:) - H2(1:R1,:)); 
   F  = [sqrt(a2)*Z2,   H0-H3,  sqrt(aY)*(Y - Tz1*W(1:R1,:)*Py')];
   G  = [sqrt(a2)*Pz2', sqrt(aT)*(H1(R1+1:R1+RA,:) - H2(R1+1:R1+RA,:)),  sqrt(aY)*W(R1+1:R1+RA,:)*Py'];
   
   if I>Jz2*Kz2
      Wz2     = Z2\F/G;
      Tz2     = Z2*Wz2;
   else
      Tz2     = F/G;
   end;
   
   T    = [Tz1 Tz2];
   Tsup = T*W;
   
   % Update Pz2 and Py;
   pinv_Tsup=pinv(Tsup);
   Py   = (pinv_Tsup*Y)';  
   Pt   = (pinv_Tsup*T)';
   
   % Update B
   XX  = permnew(Z2,I,Jz2,Kz2);
   XX1 = permnew(XX,Jz2,Kz2,I);
   XX1 = permnew(C'*XX1,RC,I,Jz2);
   XX1 = permnew(Tz2'*XX1,RA,Jz2,RC);
   H   = permnew(H,RA,RB,RC);
   [P,Ds,Q]=svd(XX1*H',0);
   B   = P*Q';
   
   % Update C
   XX  = permnew(XX,Jz2,Kz2,I);
   XX1 = permnew(XX,Kz2,I,Jz2);
   XX1 = permnew(Tz2'*XX1,RA,Jz2,Kz2);
   XX1 = permnew(B'*XX1,RB,Kz2,RA);
   H   = permnew(H,RB,RC,RA);
   [P,Ds,Q] = svd(XX1*H',0);
   C   = P*Q';
   
   % Update H 
   X1 = permnew(Z2,I,Jz2,Kz2);
   X1 = permnew(B'*X1,RB,Kz2,I);
   X1 = permnew(C'*X1,RC,I,RB);
   H  = pinv(Tz2)*X1;
   
   % Calculate Pz2
   Pz2    = permnew(H,RA,RB,RC);
   Pz2    = permnew(B*Pz2,Jz2,RC,RA);
   Pz2    = permnew(C*Pz2,Kz2,RA,Jz2);
   Pz2    = Pz2';
   
   % Update T and W. Define auxiliary matrices F and G
   F  = [sqrt(aT)*T,  sqrt(aY)*Y];
   G  = [sqrt(aT)*Pt',sqrt(aY)*Py'];
   
   if I > R1+RA
      W    = T\F/G;
      Tsup = T*W;
   else
      Tsup = F/G;
   end;
   
   % Update Pt and Py 
   pinv_Tsup = pinv(Tsup);
   Py    = (pinv_Tsup*Y)';
   Pt    = (pinv_Tsup*T)';
   
   % Calculate new f value
   lossZ1  = ssq(Z1 - Tz1*Pz1');
   lossZ2  = ssq(Z2 - Tz2*Pz2');
   lossT   = ssq(T  - Tsup*Pt');
   lossY   = ssq(Y  - Tsup*Py');
   f       = a1*lossZ1 + a2*lossZ2 + aT*lossT + aY*lossY;
   if ff < f, 
      error('loss function is increasing')
   end
   ff = f;
   
   if rem(iter,prntlag)==0
      fprintf('f= %12.8f after %g iters; lossZ1=%10.6f lossZ2=%10.6f lossY=%10.6f\n',f,iter,lossZ1,lossZ2,lossY);
   end;
   
end;
if I<Jz1,      Wz1 = pinv(Z1)*Tz1; end
if I<Jz2*Kz2,  Wz2 = pinv(Z2)*Tz2; end

switch flag
case 1
   % Explained by block scores
   % disp('explained by block scores')
   Z1dak = Tz1*Pz1';
   Z2dak = Tz2*Pz2';
   Ydak  = Tsup*Py';
   R2Z1 =(1 - ssq(Z1 - Z1dak) / ssq(Z1));
   R2Z2 =(1 - ssq(Z2 - Z2dak) / ssq(Z2));
   R2Y  =(1 - ssq(Y  - Ydak)  / ssq(Y));
   % fprintf(['R2Z1, R2Z2, R2Y ',num2str([R2Z1 R2Z2 R2Y]),'\n']);
   
case 2
   % Explained by super scores
   % disp('explained by super scores')
   P1       = (pinv(Tsup)*Z1)';
   f=1;H1_old = ones(R,RB*RC);
   
   while f>1E-8 
      % H
      X1     = permnew(Z2,I,Jz2,Kz2);
      X1     = permnew(B'*X1,RB,Kz2,I);
      X1     = permnew(C'*X1,RC,I,RB);
      H1     = pinv(Tsup)*X1;
      
      % Update B
      XX  = permnew(Z2,I,Jz2,Kz2);
      XX1 = permnew(XX,Jz2,Kz2,I);
      XX1 = permnew(C'*XX1,RC,I,Jz2);
      XX1 = permnew(Tsup'*XX1,R,Jz2,RC);
      H1  = permnew(H1,R,RB,RC);
      [P,Ds,Q]=svd(XX1*H1',0);
      B   = P*Q';
      
      % Update C
      XX  = permnew(XX,Jz2,Kz2,I);
      XX1 = permnew(XX,Kz2,I,Jz2);
      XX1 = permnew(Tsup'*XX1,R,Jz2,Kz2);
      XX1 = permnew(B'*XX1,RB,Kz2,R);
      H1  = permnew(H1,RB,RC,R);
      [P,Ds,Q] = svd(XX1*H1',0);
      C   = P*Q';
      
      % Update H 
      X1 = permnew(Z2,I,Jz2,Kz2);
      X1 = permnew(B'*X1,RB,Kz2,I);
      X1 = permnew(C'*X1,RC,I,RB);
      H1  = pinv(Tsup)*X1;
      
      f=ssq(H1-H1_old)/ssq(H1);
      H1_old=H1;
   end
   
   % Calculate P2
   P2    = permnew(H1,R,RB,RC);
   P2    = permnew(B*P2,Jz2,RC,R);
   P2    = permnew(C*P2,Kz2,R,Jz2);
   P2    = P2';
   
   Z1dak    = Tsup*P1';
   Z2dak    = Tsup*P2';
   Ydak     = Tsup*Py';
   R2Z1     = (1 - ssq(Z1 - Z1dak) / ssq(Z1));
   R2Z2     = (1 - ssq(Z2 - Z2dak) / ssq(Z2));
   R2Y      = (1 - ssq(Y  - Ydak)  / ssq(Y));
   % fprintf(['R2Z1, R2Z2, R2Y ',num2str([R2Z1 R2Z2 R2Y]),'\n']);
end
