function [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,Pt,Py,R2Z1,R2Z2,R2Y] = mbtw(Z1,Z2,Y,Rz1,Rz2,R,k1,k2,c,flag,conv);
%        [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,Pt,Py,R2Z1,R2Z2,R2Y] = mbtw(Z1,Z2,Y,Rz1,Rz2,R,k1,k2,c,flag,conv);
% 
% Two-way multiblock covariates regression. Matrices Z1 and Z2 predict Y.
% Rz1, Rz2 and R can be selected different.
% After convergence all score matrices are orthogonalized.
% 
% Minimize f(Wz1,Wz2,W)= a1||Z1-Z1Wz1Pz1'||^2 + 
%                        a2||Z2-Z2Wz2Pz2'||^2 +
%                        aT||T-TWPt'||^2  +
%                        aY||Y-TWPy'||^2
%
% Input: Z1      (I x Jz1)
%        Z2      (I x Jz2)
%        Y       (I x Jy)
%        Rz1:    number of components of block Z1 (set Rz1=0 if it should not be used)
%        Rz2:    number of components of block Z2
%        R:      number of final components
%        k1:     weight of Z1-block
%        k2:     weight of Z2-block
%        c:      gamma value, alphaT = c*alphaY
%        flag:   Calculate R2 values using 2(superscores) 1(blockscores)
%        conv:   convergence criterion (usually 1E-6)
%
% Output: Wz1:   weight vectors for Z1 (Jz1 x Rz1)
%         Wz2:   weight vectors for Z2 (Jz2 x Rz2)
%         W  :   weight vectors for Tsup (Rz1+Rz2 x R)
%         Tz1:   scores of Z1 (I x Jz1)   T = [Tz1  Tz2]   (I x Rz1+Rz2)
%         Tz2:   scores of Z2 (I x Jz2)
%         Tsup:  scores to predict Y (I x R)
%         Pz1:   loadings of Z (Rz1 x Jz1)
%         Pz2:   loadings of X (Rz2 x Jz2)
%         Pt:    loadings of T (Rz1+Rz2 x R)
%         Py:    loadings of Y (R x Jy)
%         R2Z, R2X, R2Y: explained percentage of variation in the blocks
%
% Uses: ssq.m, ed.m
%
% Johan Westerhuis, 21/6/99 (westerhuis@its.chem.uva.nl)
%   ==========================================================================
%   Copyright 2005 Biosystems Data Analysis Group ; Universiteit van Amsterdam
%   ==========================================================================

% Revised 10-9-1999; JAW
% alpha T is changed from (1-aY) to c*aY with constraint a1+a2+aT+aY=1

prntlag = 1000;

% Sizes of matrices
[I,Jz1] = size(Z1);
[I,Jz2] = size(Z2);
[I,Jy]  = size(Y);

% Define alpha values
aY = 1/(1+k1+k2+c); %aY + caY + k1aY + k2aY = 1
aT = c*aY;          %1  + c   + k1   + k2   = 1/aY
a1 = k1*aY;
a2 = k2*aY;

% Calculate starting values of Wz1, Wz2, and W
% First Wz1
if I>Jz1
   [P,D,Q]  = svd(Z1,0);
   Wz1  = Q(:,1:Rz1)/D(1:Rz1,1:Rz1);
else
   [P,D]    = ed(Z1*Z1');
end;
Tz1 = P(:,1:Rz1);

% Now Wz2
if I>Jz2
   [P,D,Q]  = svd(Z2,0);
   Wz2  = Q(:,1:Rz2)/D(1:Rz2,1:Rz2);
else
   [P,D]    = ed(Z2*Z2');
end;
Tz2 = P(:,1:Rz2);

T   = [Tz1 Tz2];

% And W
if I    > Rz1+Rz2
   [P,D,Q]  = svd(T,0);
   W    = Q(:,1:R)/D(1:R,1:R);
else
   [P,D]    = ed(T*T');
end;
Tsup    = P(:,1:R);

% Calculate starting values of Pz1, Pz2 and Py
Pz1 = (pinv(Tz1)*Z1)';  % at start all T's are orthonormal
Pz2 = (pinv(Tz2)*Z2)';  
Py  = (pinv(Tsup)*Y)';
Pt    = (pinv(Tsup)*T)';

% Calculate start loss functions
lossZ1 = ssq(Z1 - Tz1*Pz1');
lossZ2 = ssq(Z2 - Tz2*Pz2');
lossT  = ssq(T  - Tsup*Pt');
lossY  = ssq(Y  - Tsup*Py');
f      = a1*lossZ1 + a2*lossZ2 + aT*lossT+aY*lossY;
ff     = f;
% fprintf('function value at start is %12.8f \n',f);

if I    < Jz1,    Wz1   = pinv(Z1)*Tz1; end
if I    < Jz2,    Wz2   = pinv(Z2)*Tz2; end
if I    < Rz1+Rz2, W    = pinv(T)*Tsup; end 

% Start iterations
fold    = f+2*conv*f;
iter    = 0;

H0    = zeros(I,Rz1+Rz2);
H1    = eye(Rz1+Rz2);

while fold-f>conv*f | iter<2
   fold   = f;
   iter   = iter+1;
   
   % Update Tz1 and Wz1. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(aT) * Tz2*(H1(Rz1+1:Rz1+Rz2,:) - H2(Rz1+1:Rz1+Rz2,:));
   F    = [sqrt(a1)*Z1,  H0-H3, sqrt(aY)*(Y - Tz2*W(Rz1+1:Rz1+Rz2,:)*Py')];
   G    = [sqrt(a1)*Pz1',sqrt(aT)*(H1(1:Rz1,:) - H2(1:Rz1,:)),  sqrt(aY)*W(1:Rz1,:)*Py'];
   
   if I > Jz1
      Wz1   = Z1\F/G;
      Tz1   = Z1*Wz1;
   else
      Tz1   = F/G;
   end;
   
   T    = [Tz1 Tz2];
   Tsup = T*W;
   
   % Update Pz1, Pt and Py
   Pz1  =(pinv(Tz1)*Z1)';
   pinv_Tsup = pinv(Tsup);
   Py   =(pinv_Tsup*Y)';
   Pt   =(pinv_Tsup*T)';
   
   % Update Tz2 and Wz2. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(aT) * Tz1*(H1(1:Rz1,:) - H2(1:Rz1,:)); 
   F    = [sqrt(a2)*Z2,   H0-H3,  sqrt(aY)*(Y - Tz1*W(1:Rz1,:)*Py')];
   G    = [sqrt(a2)*Pz2', sqrt(aT)*(H1(Rz1+1:Rz1+Rz2,:) - H2(Rz1+1:Rz1+Rz2,:)),  sqrt(aY)*W(Rz1+1:Rz1+Rz2,:)*Py'];
   
   if I  > Jz2
      Wz2   = Z2\F/G;
      Tz2   = Z2*Wz2;
   else
      Tz2   = F/G;
   end;
   
   T    = [Tz1 Tz2];
   Tsup = T*W;
   
   % Update Pz2, Pt and Py
   Pz2       = (pinv(Tz2)*Z2)';
   pinv_Tsup = pinv(Tsup);
   Py        = (pinv_Tsup*Y)';
   Pt        = (pinv_Tsup*T)';
   
   % Update T and W. Define auxiliary matrices F and G
   F  = [sqrt(aT)*T,  sqrt(aY)*Y];
   G  = [sqrt(aT)*Pt', sqrt(aY)*Py'];
   
   if I > Rz1+Rz2
      W    = T\F/G;
      Tsup = T*W;
   else
      Tsup = F/G;
   end;
   
   % Update Pt and Py 
   pinv_Tsup = pinv(Tsup);
   Py        =  (pinv_Tsup*Y)';
   Pt        = (pinv_Tsup*T)';
   
   % Calculate new f value
   lossZ1    = ssq(Z1 - Tz1*Pz1');
   lossZ2    = ssq(Z2 - Tz2*Pz2');
   lossT     = ssq(T  - Tsup*Pt');
   lossY     = ssq(Y  - Tsup*Py');
   f        = a1*lossZ1 + a2*lossZ2 + aT*lossT + aY*lossY;
   if ff < f, 
      error('loss function is increasing')
   end
   ff = f;
   
   if rem(iter,prntlag)==0
      fprintf('f= %12.8f after %g iters; lossZ=%10.6f lossX=%10.6f lossY=%10.6f\n',f,iter,lossZ1,lossZ2,lossY);
   end;
end;

% Orthogonalizing Tz1 and Tz2;
[Tz1,qr1]  = qr(Tz1,0);
Wz1        = Wz1 * inv(qr1);
Pz1        = Pz1 * qr1';
W(1:Rz1,:) = qr1 * W(1:Rz1,:);

[Tz2,qr2]  = qr(Tz2,0);
Wz2      = Wz2 * inv(qr2);
Pz2        = Pz2 * qr2';
W(Rz1+1:Rz1+Rz2,:) = qr2 * W(Rz1+1:Rz1+Rz2,:);

% Orthogonalization of Tsup
[Tsup,qrsup]  = qr(Tsup,0);
W             = W * inv(qrsup);
Py            = Py * qrsup';
Pt            = Pt * qrsup';

if I    < Jz1,    Wz1   = pinv(Z1)*Tz1; end
if I    < Jz2,    Wz2   = pinv(Z2)*Tz2; end
if I    < Rz1+Rz2, W    = pinv(T)*Tsup; end 

% Give percentage explained using blockscores(1) or superscores(2)
switch flag
case 1
   % explained by block scores
   R2Z1 =(1 - ssq(Z1 - Tz1*Pz1') / ssq(Z1));
   R2Z2 =(1 - ssq(Z2 - Tz2*Pz2') / ssq(Z2));
   R2Y  =(1 - ssq(Y  - Tsup*Py') / ssq(Y));
      
case 2
   % Explained by super scores
   P1   = (pinv(Tsup)*Z1)';
   P2      = (pinv(Tsup)*Z2)';
   R2Z1 =(1 - ssq(Z1 - Tsup*P1') / ssq(Z1));
   R2Z2 =(1 - ssq(Z2 - Tsup*P2') / ssq(Z2));
   R2Y  =(1 - ssq(Y  - Tsup*Py') / ssq(Y));
end