function [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,Pt,B,C,H,R2Z1,R2Z2]=mwmbcomp(Z1,Z2,Jz2,Kz2,R1,RA,RB,RC,R,a1,a2,conv);
% function [Wz1,Wz2,W,Tz1,Tz2,Tsup,Pz1,Pz2,Pt,B,C,H,R2Z1,R2Z2]=mwmbcomp(Z1,Z2,Jz2,Kz2,R1,RA,RB,RC,R,a1,a2,conv);
% 
% Multiway multiblock component model for two way matrix Z1 and multiway array Z2.
% A Tucker 3 structure is applied on Z2.
% The number of components can be chosen differently for the predictor blocks.
%  
% Minimize f(Wz1,Wz2)  = a1||Z1-Z1Wz1Pz1'||^2 + 
%                        a2||Z2-Z2Wz2H(C'kronB')||^2 +
%                        a3||T-TWPt'||^2 +
%
% Input: 
%    Z1(IxJz1)          Predictor block 1
%    Z2(Ix(Jz2*Kz2))    Predictor block 2 unfolded with batch direction maintained
%    Jz2                number of variables in process variables direction
%    Kz2                number of variables in time direction
%    R1                 number of components for Z1
%    RA                 number of components for Z2 in the batch direction
%    RB                 number of components for Z2 in the process variable direction
%    RC                 number of components for Z2 in the time direction
%    R                  number of final components
%    a1                 weight of Z1-block (between 0 and 1)
%    a2                 weight of Z2-block (between 0 and 1)
%    conv               convergence criterion 
%
% Output: 
%    Wz1(Jz1xR1)        weight vectors for Z1
%    Wz2((Jz2*Kz2)xRA)  weight vectors for Z2
%    W((R1+RA)xR        weight vectors for T)
%    Tz1(IxR1)          block scores of Z1
%    Tz2(IxRA)          block scores of Z2
%    Tsup(IxR)          super scores to predict Y
%    Pz1(Jz1xR1)        loadings of Z1
%    Pz2((Jz2xKz2)xRA)  loadings of Z2 (Pz2=H(C'kronB'))
%    Pt((R1+RA)xR)
%    B(Jz2xRB)          loadings of Z2 in process variables direction
%    C(Kz2xRC)          loadings of Z2 in time direction
%    H(Rx(RB*RC))       core array
%    R2Z1               explained variation of Y
%    R2Z2               explained variation of Y
%
% Other variables:
%    lossZ1             parts of the total loss function
%    lossZ2
%    lossT
%    f

% Uses: ssq.m, ed.m, permnew.m
%
% Johan Westerhuis, 21/10/99 (westerhuis@its.chem.uva.nl)
%   ==========================================================================
%   Copyright 2005 Biosystems Data Analysis Group ; Universiteit van Amsterdam
%   ==========================================================================

prntlag = 50;

% Sizes of matrices
[I,Jz1] = size(Z1);
[I,JK]  = size(Z2);   % JK = Jz2*Kz2

% Define a3
a3=1-a1-a2;

% Calculate starting values of Wz1, Wz2
% First Wz1
if I>Jz1
   [P,D,Q]   = svd(Z1,0);
   Wz1       = Q(:,1:R1)/D(1:R1,1:R1);
else
   [P,D]     = ed(Z1*Z1');
end;
Tz1      = P(:,1:R1);

% Now Wz2
if I > Jz2*Kz2
   [P,D,Q]   = svd(Z2,0);
   Wz2      = Q(:,1:RA)/D(1:RA,1:RA);
else
   [P,D]     = ed(Z2*Z2');
end;
Tz2      = P(:,1:RA);

T        = [Tz1 Tz2];

% And W
if I    > R1+RA
   [P,D,Q]   = svd(T,0);
   W         = Q(:,1:R)/D(1:R,1:R);
else
   [P,D]     = ed(T*T');
end;
Tsup       = P(:,1:R);

% Update Pz1, Pz2 and Py;
Pz1   = (pinv(Tz1)*Z1)';
Pt    = (pinv(Tsup)*T)';

% Calculate starting values for B, C and H
% B
XX = permnew(Z2,I,Jz2,Kz2);
if Jz2 > I*Kz2
   [P,Dd,Q] = svd(XX,0);
else
   [P,Dd]   = ed(XX*XX');
end;
B = P(:,1:RB);

% C
XX = permnew(Z2,Jz2,Kz2,I);
if Kz2 > I*Jz2
   [P,Dd,Q] = svd(XX,0);
else
   [P,Dd]   = ed(XX*XX');
end;
C = P(:,1:RC);

% H
X1     = permnew(Z2,I,Jz2,Kz2);
X1     = permnew(B'*X1,RB,Kz2,I);
X1     = permnew(C'*X1,RC,I,RB);
H      = pinv(Tz2)*X1;

% Calculate Pz2
Pz2    = permnew(H,RA,RB,RC);
Pz2    = permnew(B*Pz2,Jz2,RC,RA);
Pz2    = permnew(C*Pz2,Kz2,RA,Jz2);
Pz2    = Pz2';

% Calculate start loss functions
lossZ1 = ssq(Z1 - Tz1*Pz1');
lossZ2 = ssq(Z2 - Tz2*Pz2');
lossT  = ssq(T  - Tsup*Pt');
f      = a1*lossZ1 + a2*lossZ2 + a3*lossT;
ff     = f;

% Calculate weight matrices
if I    < Jz1,    Wz1   = pinv(Z1)*Tz1; end
if I    < Jz2,    Wz2   = pinv(Z2)*Tz2; end
if I    < R1+RA,   W    = pinv(T)*Tsup; end 

% Start iterations
fold       = f + 2*conv*f;
iter       = 0;

H0  = zeros(I,R1+RA);
H1  = eye(R1+RA);

while fold-f>conv*f | iter<2
   fold   = f;
   iter   = iter+1;
   
   % Update Tz1 and Wz1. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(a3) * Tz2*(H1(R1+1:R1+RA,:) - H2(R1+1:R1+RA,:));
   F  = [sqrt(a1)*Z1,  H0-H3];
   G  = [sqrt(a1)*Pz1',sqrt(a3)*(H1(1:R1,:) - H2(1:R1,:))];

   if I > Jz1
      Wz1     = Z1\F/G;
      Tz1     = Z1*Wz1;
   else
      Tz1     = F/G;
   end;
   
   % Update T, Tsup, Pz1 and Pt
   T    = [Tz1 Tz2];
   Tsup = T*W;
   
   Pz1  =(pinv(Tz1)*Z1)';
   pinv_Tsup = pinv(Tsup);
   Pt   =(pinv_Tsup*T)';
   
   % Update Tz2 and Wz2. Define auxiliary matrices F and G
   H2 = W*Pt';
   H3 = sqrt(a3) * Tz1*(H1(1:R1,:) - H2(1:R1,:)); 
   F  = [sqrt(a2)*Z2,   H0-H3];
   G  = [sqrt(a2)*Pz2', sqrt(a3)*(H1(R1+1:R1+RA,:) - H2(R1+1:R1+RA,:))];
   
   if I>Jz2*Kz2
      Wz2     = Z2\F/G;
      Tz2     = Z2*Wz2;
   else
      Tz2     = F/G;
   end;
   
   % Update T, Tsup, Pt and Pz2(H(C'kronB'))
   T    = [Tz1 Tz2];
   Tsup = T*W;

   pinv_Tsup=pinv(Tsup);
   Pt   = (pinv_Tsup*T)';
   
   % Update B
   XX  = permnew(Z2,I,Jz2,Kz2);
   XX1 = permnew(XX,Jz2,Kz2,I);
   XX1 = permnew(C'*XX1,RC,I,Jz2);
   XX1 = permnew(Tz2'*XX1,RA,Jz2,RC);
   H   = permnew(H,RA,RB,RC);
   [P,Ds,Q] = svd(XX1*H',0);
   B   = P*Q';

   % Update C
   XX  = permnew(XX,Jz2,Kz2,I);
   XX1 = permnew(XX,Kz2,I,Jz2);
   XX1 = permnew(Tz2'*XX1,RA,Jz2,Kz2);
   XX1 = permnew(B'*XX1,RB,Kz2,RA);
   H   = permnew(H,RB,RC,RA);
   [P,Ds,Q] = svd(XX1*H',0);
   C   = P*Q';

   % Update H 
   X1 = permnew(Z2,I,Jz2,Kz2);
   X1 = permnew(B'*X1,RB,Kz2,I);
   X1 = permnew(C'*X1,RC,I,RB);
   H  = pinv(Tz2)*X1;
   
   % Calculate Pz2
   Pz2    = permnew(H,RA,RB,RC);
   Pz2    = permnew(B*Pz2,Jz2,RC,RA);
   Pz2    = permnew(C*Pz2,Kz2,RA,Jz2);
   Pz2    = Pz2';
   
   % Update T and W for the loss function: minW ||T-TWPt||^2
   W     = pinv(Pt');
   Tsup  = T*W;
   
   % Update Pt
   pinv_Tsup = pinv(Tsup);
   Pt    = (pinv_Tsup*T)';
   
   % Calculate new f value
   lossZ1  = ssq(Z1 - Tz1*Pz1');
   lossZ2  = ssq(Z2 - Tz2*Pz2');
   lossT   = ssq(T  - Tsup*Pt');
   f       = a1*lossZ1 + a2*lossZ2 + a3*lossT;
   if ff < f, 
         error('loss function is increasing')
   end
   ff = f;

   if rem(iter,prntlag)==0
      fprintf('f= %12.8f after %g iters; lossZ1=%10.6f lossZ2=%10.6f\n',f,iter,lossZ1,lossZ2);
   end;
   
end;
if I<Jz1,      Wz1 = pinv(Z1)*Tz1; end
if I<Jz2*Kz2,  Wz2 = pinv(Z2)*Tz2; end


% Explained by block scores
% disp('explained by block scores')
Z1dak = Tz1*Pz1';
Z2dak = Tz2*Pz2';
R2Z1    =(1 - ssq(Z1 - Z1dak) / ssq(Z1));
R2Z2    =(1 - ssq(Z2 - Z2dak) / ssq(Z2));
% fprintf(['R2Z1, R2Z2 ',num2str([R2Z1 R2Z2]),'\n']);
   
% Explained by super scores
% disp('explained by super scores')
P1       = (pinv(Tsup)*Z1)';
f=1;H1_old = ones(R,RB*RC);

while f>1E-8 
   % H
   X1     = permnew(Z2,I,Jz2,Kz2);
   X1     = permnew(B'*X1,RB,Kz2,I);
   X1     = permnew(C'*X1,RC,I,RB);
   H1     = pinv(Tsup)*X1;

   % Update B
   XX  = permnew(Z2,I,Jz2,Kz2);
   XX1 = permnew(XX,Jz2,Kz2,I);
   XX1 = permnew(C'*XX1,RC,I,Jz2);
   XX1 = permnew(Tsup'*XX1,R,Jz2,RC);
   H1  = permnew(H1,R,RB,RC);
   [P,Ds,Q]=svd(XX1*H1',0);
   B   = P*Q';

   % Update C
   XX  = permnew(XX,Jz2,Kz2,I);
   XX1 = permnew(XX,Kz2,I,Jz2);
   XX1 = permnew(Tsup'*XX1,R,Jz2,Kz2);
   XX1 = permnew(B'*XX1,RB,Kz2,R);
   H1  = permnew(H1,RB,RC,R);
   [P,Ds,Q] = svd(XX1*H1',0);
   C   = P*Q';

   % Update H 
   X1 = permnew(Z2,I,Jz2,Kz2);
   X1 = permnew(B'*X1,RB,Kz2,I);
   X1 = permnew(C'*X1,RC,I,RB);
   H1  = pinv(Tsup)*X1;

   f=ssq(H1-H1_old)/ssq(H1);
   H1_old=H1;
end

% Calculate P2
P2    = permnew(H1,R,RB,RC);
P2    = permnew(B*P2,Jz2,RC,R);
P2    = permnew(C*P2,Kz2,R,Jz2);
P2    = P2';

Z1dak    = Tsup*P1';
Z2dak    = Tsup*P2';
R2Z1     = (1 - ssq(Z1 - Z1dak) / ssq(Z1));
R2Z2     = (1 - ssq(Z2 - Z2dak) / ssq(Z2));
% fprintf(['R2Z1, R2Z2 ',num2str([R2Z1 R2Z2]),'\n']);
